<?php
declare(strict_types=1);
/*
 * Your installation or use of this SugarCRM file is subject to the applicable
 * terms available at
 * http://support.sugarcrm.com/Resources/Master_Subscription_Agreements/.
 * If you do not agree to all of the applicable terms or do not have the
 * authority to bind the entity as an authorized representative, then do not
 * install or use this SugarCRM file.
 *
 * Copyright (C) SugarCRM Inc. All rights reserved.
 */

namespace Sugarcrm\Sugarcrm\Security\ModuleScanner;

final class SweetShield
{
    public const SWEET_PREFIX = 'sooo_sweeet_';
    public const SWEET_INTERFACE = '\\SoooSweeet';
    private static array $denyLists;

    public static function validPath(string $filename): string
    {
        if (!check_file_name($filename)) {
            throw new \RuntimeException("Trying to include invalid file: $filename");
        }
        return $filename;
    }

    public static function callMethod($object, string $method, ...$params)
    {
        if (!self::isAllowedMethod($object, $method)) { // checks against allowlist
            $error = sprintf('Code attempted to call denylisted method "%s::%s"', is_object($object) ? get_class($object) : (string)$object, $method);
            throw new \RuntimeException($error);
        }
        return call_user_func_array([$object, $method], $params);
    }

    public static function callFunction(string $function, ...$params)
    {
        if (!self::isAllowedFunction($function)) {// checks against allowlist
            if (is_callable(['self', 'safe_' . $function])) {// safe alternative for internal function
                return call_user_func_array(['self', 'safe_' . $function], $params);
            } elseif (!in_array(strtolower(self::SWEET_PREFIX . $function), get_defined_functions()['user'])) {
                $error = sprintf('Code attempted to call denylisted function "%s"', $function);
                throw new \RuntimeException($error);
            } else {
                return call_user_func_array(self::SWEET_PREFIX . $function, $params);
            }
        }
        return call_user_func_array($function, $params);
    }

    /**
     * $allowedFunctions in initial variant are:
     * $internal = get_defined_functions()['internal'];
     * $allowedFunctions = array_diff($internal, $denylist, $known_denylist_bypasses);
     */
    public static function isAllowedFunction(string $function): bool
    {
        if (self::isInternalFunction($function)) {
            $deniedFunctions = self::getDenyList('functions');
            return !in_array(strtolower($function), $deniedFunctions);
        }
        if (0 === strpos($function, self::SWEET_PREFIX)) {
            return true;
        }
        return false;
    }

    public static function isInternalFunction(string $function): bool
    {
        return in_array($function, get_defined_functions()['internal']);
    }

    public static function isAllowedMethod($obj, string $method): bool
    {
        $class = is_object($obj) ? get_class($obj) : (string)$obj;

        // Allow all classes defined in MLP
        if (is_subclass_of($class, self::SWEET_INTERFACE)) {
            return true;
        }
        // Allow classes that extend SugarBean
        if (is_subclass_of($class, \SugarBean::class)) {
            return true;
        }

        if (in_array($class, self::getDenyList('classes'))) {
            return false;
        }

        $deniedMethods = self::getDenyList('methods');
        if (in_array($method, $deniedMethods)) {//denied for all classes
            return false;
        }

        if (isset($deniedMethods[$method]) && $deniedMethods[$method] === $class) {
            return false;
        }
        return false;
    }

    /**
     * Safe variant of unserialize, enforces ['allowed_classes' => false] to be passed as an option
     * @param string $data
     * @param array $options
     * @return mixed
     */
    private static function safe_unserialize(string $data, array $options = [])
    {
        $options['allowed_classes'] = false;
        return unserialize($data, $options);
    }


    private static function safe_file_get_contents(string $filename, bool $use_include_path = false, $context = null, int $offset = 0, ?int $length = null)
    {
        if (substr($filename, 0, 9) !== 'upload://') {
            $filename = "upload://$filename";
        }
        //Disable include_path and ignore $context
        return file_get_contents($filename, false, null, $offset, $length);
    }

    private static function safe_file_put_contents(string $filename, mixed $data, int $flags = 0, $context = null)
    {
        if (substr($filename, 0, 9) !== 'upload://') {
            $filename = "upload://$filename";
        }
        //ignore $context
        return file_put_contents($filename, $data, $flags);
    }

    private static function getDenyList(string $listName): array
    {
        if (!isset(self::$denyLists)) {
            self::$denyLists = (new \ModuleScanner())->getEffectiveDenyLists();
        }
        return self::$denyLists[$listName];
    }
}
