<?php
declare(strict_types=1);
/*
 * Your installation or use of this SugarCRM file is subject to the applicable
 * terms available at
 * http://support.sugarcrm.com/Resources/Master_Subscription_Agreements/.
 * If you do not agree to all of the applicable terms or do not have the
 * authority to bind the entity as an authorized representative, then do not
 * install or use this SugarCRM file.
 *
 * Copyright (C) SugarCRM Inc. All rights reserved.
 */

namespace Sugarcrm\Sugarcrm\Security\ModuleScanner;

use PhpParser\BuilderFactory;
use PhpParser\Node;
use PhpParser\Node\Expr\Variable;
use PhpParser\NodeVisitorAbstract;

final class TranslateVisitor extends NodeVisitorAbstract
{
    /**
     * @var Node\Stmt\Function_[]
     */
    private array $customFunctions = [];

    public function leaveNode(Node $node)
    {
        if ($node instanceof Node\Expr\MethodCall) {
            $factory = new BuilderFactory;
            if ($node->name instanceof Node\Identifier) {
                $methodName = $node->name->name;
            } else {
                $methodName = $node->name;
            }
            $object = $node->var;
            if ($object instanceof Variable && (string)$object->name === 'this') {
                return $node;
            }

            $methodArgs = $node->args;
            array_unshift($methodArgs, $object, $methodName);
            return $factory->staticCall('\\' . SweetShield::class, 'callMethod', $methodArgs);
        } elseif ($node instanceof Node\Expr\FuncCall) {
            $funcName = $node->name instanceof Node\Name ? $node->name->toString() : $node->name;
            /**
             * Special case for 'unserialize'. Enforce the second param ['allowed_classes' => false]
             * to prevent PHP Object injection attacks
             */
            if ($funcName === 'unserialize') {
                $node->args[1] = new Node\Arg(new Node\Expr\Array_(
                    [
                        new Node\Expr\ArrayItem(
                            new Node\Expr\ConstFetch(new Node\Name('false')),
                            new Node\Scalar\String_('allowed_classes')
                        ),
                    ]
                ));
                return $node;
            } else {
                $factory = new BuilderFactory;
                // Do not wrap allowed functions with callFunction()

                if ($node->name instanceof Node\Name && SweetShield::isAllowedFunction($node->name->toString())) {
                    return $node;
                }
                $funcArgs = $node->args;
                array_unshift($funcArgs, $funcName);
                return $factory->staticCall('\\' . SweetShield::class, 'callFunction', $funcArgs);
            }
        } elseif ($node instanceof Node\Expr\StaticCall) {
            $class = $node->class;
            $method = $node->name;

            /**
             * Taking care only about dynamically named classes and methods, statically resolvable cases
             * should be handled in \Sugarcrm\Sugarcrm\Security\ModuleScanner\BlacklistVisitor
             */
            if ($this->isNameDynamicallyCalled($class) || $this->isNameDynamicallyCalled($method)) {
                $factory = new BuilderFactory;
                if ($class instanceof Node\Name && in_array($class->toString(), ['self', 'static'], true)) {
                    return $node;
                }
                if ($node->name instanceof Node\Identifier) {
                    $methodName = $node->name->name;
                } else {
                    $methodName = $node->name;
                }
                $methodArgs = $node->args;
                array_unshift($methodArgs, $class, $methodName);
                return $factory->staticCall('\\' . SweetShield::class, 'callMethod', $methodArgs);
            }
        } elseif ($node instanceof Node\Expr\Include_) {
            $factory = new BuilderFactory();
            /**
             * Validate filename passed to include/require to prevent Path Traversal attacks
             * and abusing of stream filters and stream wrappers
             */
            $node->expr = $factory->staticCall('\\' . SweetShield::class, 'validPath', [$node->expr]);
            return $node;
        }
    }

    public function enterNode(Node $node)
    {
        if ($node instanceof Node\Stmt\Function_) {
            $parent = $node->getAttribute('parent');
            $factory = new BuilderFactory();
            /**
             * Creating a proxy function with the original name. Attributes and return type are ignored at the moment
             */
            $proxy = $factory->function((string)$node->name)
                ->addParams($node->getParams())
                ->addStmt(new Node\Stmt\Return_(
                    $factory->funcCall(
                        SweetShield::SWEET_PREFIX . $node->name,
                        array_map(function (Node\Param $param) {
                            return $param->var;
                        }, $node->getParams())
                    )
                ))
                ->getNode();

            if ($parent) {
                /**
                 * We are inside the namespace
                 */
                $parent->setAttribute('customFunctions', array_merge($parent->getAttribute('customFunctions', []), [$proxy]));
            } else {
                $this->customFunctions[] = $proxy;
            }
            /**
             * Rename the original function
             */
            $node->name = SweetShield::SWEET_PREFIX . $node->name;
            return $node;
        }
        if ($node instanceof Node\Stmt\Class_) {
            $sweetInterface = [new Node\Name(SweetShield::SWEET_INTERFACE)];
            $node->implements = $node->implements ? array_merge($node->implements, $sweetInterface) : $sweetInterface;
        }
    }

    private function isNameDynamicallyCalled($name): bool
    {
        if ($name instanceof Node\Expr\Variable ||
            $name instanceof Node\Expr\ConstFetch ||
            $name instanceof Node\Expr\FuncCall ||
            $name instanceof Node\Expr\PropertyFetch ||
            $name instanceof Node\Expr\ClassConstFetch ||
            $name instanceof Node\Expr\MethodCall ||
            $name instanceof Node\Expr\StaticCall ||
            $name instanceof Node\Expr\BinaryOp\Concat ||
            $name instanceof Node\Expr\Cast\String_ ||
            $name instanceof Node\Scalar\String_ ||
            $name instanceof Node\Expr\ArrayDimFetch ||
            $name instanceof Node\Scalar\Encapsed
        ) {
            return true;
        }

        return false;
    }

    /**
     * Add proxy functions with original function names for all custom functions in the script to prevent BC breaks
     * @param array $nodes
     * @return array|null
     */
    public function afterTraverse(array $nodes)
    {
        // functions in the namespace
        foreach ($nodes as $node) {
            if ($functions = $node->getAttribute('customFunctions')) {
                foreach ($functions as $function) {
                    $node->stmts[] = $function;
                }
            }
        }
        // functions in the global scope
        foreach ($this->customFunctions as $function) {
            $nodes[] = $function;
        }
        return $nodes;
    }
}
